#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# These functions provide a way to save and load Macal compiled "bytecode".

from typing import Any
import pickle

MAGIC_BYTES: bytes = b'BMC'
FILE_FORMAT_VERSION: int = 5
BINARY_FILE_EXTENSION: str = '.mcb'
SOURCE_FILE_EXTENSION: str = '.mcl'

def SaveBinary(filename: str, data: Any) -> None:
    if not filename.endswith(BINARY_FILE_EXTENSION):
        filename += BINARY_FILE_EXTENSION
    with open(filename, 'wb') as f:
        f.write(MAGIC_BYTES)
        f.write(FILE_FORMAT_VERSION.to_bytes(4, byteorder='little'))
        code = pickle.dumps(data)
        code_size = len(code)
        f.write(code_size.to_bytes(4, byteorder='little'))
        f.write(code)

def LoadBinary(filename: str) -> Any:
    if not filename.endswith(BINARY_FILE_EXTENSION):
        filename += BINARY_FILE_EXTENSION
    with open(filename, 'rb') as f:
        magic = f.read(3)
        if magic != MAGIC_BYTES:
            raise Exception(f'Invalid magic bytes: {magic}')
        version = int.from_bytes(f.read(4), byteorder='little')
        if version != FILE_FORMAT_VERSION:
            raise Exception(f'Invalid file format version: {version}')
        code_size = int.from_bytes(f.read(4), byteorder='little')
        code = f.read(code_size)
        return pickle.loads(code)



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the ast node for the debug Stack statement.

from __future__ import annotations

from typing import Optional

from .ast_node_expression import AstNodeExpression
from .ast_node_statement import AstNodeStatement
from .ast_nodetype import AstNodetype
from .lex_token import LexToken


class AstNodeDebugStackStatement(AstNodeStatement):
    def __init__(self, token: LexToken) -> AstNodeDebugStackStatement:
        super().__init__(token)
        self.expr_type = AstNodetype.DEBUG_STACK_STATEMENT

    def __str__(self) -> str:
        return f"DebugStackStatement()"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = "") -> None:
        print(f"{indent}DebugStackStatement()")



# Product:   Macal
# Author:    Marco Caspers
# Date:      16-10-2023
#
#    This library is licensed under the MIT license.
#
#    (c) 2023 Westcon-Comstor
#    (c) 2023 WestconGroup, Inc.
#    (c) 2023 WestconGroup International Limited
#    (c) 2023 WestconGroup EMEA Operations Limited
#    (c) 2023 WestconGroup European Operations Limited
#    (c) 2023 Sama Development Team
#
# System Library external functions
#

import os
import platform
import sys
from typing import Any


def Console(*args, **kwargs):
    print(*args, **kwargs)


def Array(*args) -> None:
    arr = []
    for arg in args:
        arr.append(arg)
    return arr


def RecordHasField(record: dict, fieldname: str) -> bool:
    """Checks if a record has a field"""
    if not isinstance(record, dict):
        raise TypeError("RecordHasField can only be used on a record")
    if not isinstance(fieldname, str):
        raise TypeError("RecordHasField can only be used with a string as fieldname")
    return fieldname in record


def GetPlatform() -> str:
    return platform.system()


def RecordItems(var: dict) -> list:
    """Implementation of Items function used in conjunction with foreach for iterating over records.  Items returns key/value pairs."""
    if not isinstance(var, dict):
        raise TypeError("Items can only be used on a record")
    return [{key: value} for key, value in var.items()]


def RecordItemKey(var: dict) -> str:
    """Implementation of Key function used in conjunction the Items function that returns key/value pairs. Key returns the key part of a key value pair."""
    if not isinstance(var, dict):
        raise TypeError("ItemKey can only be used on a record")
    for k, _ in var.items():  # there are different ways, but this is by far the most simple and safe way to do it.
        return k


def RecordKeys(var: dict) -> list:
    """Implementation of Keys function used in conjunction the Items function that returns key/value pairs. Key returns the key part of a key value pair."""
    if not isinstance(var, dict):
        raise TypeError("Keys can only be used on a record")
    return [k for k in var.keys()]


def RecordItemValue(var: dict) -> Any:
    """Implementation of Value function used in conjunction the Items function that returns key/value pairs. Value returns the value part of a key value pair."""
    if not isinstance(var, dict):
        raise TypeError("ItemValue can only be used on a record")
    for _, v in var.items():  # there are different ways, but this is by far the most simple and safe way to do it.
        return v


def RecordValues(var: dict) -> list:
    """Implementation of Value function used in conjunction the Items function that returns key/value pairs. Value returns the value part of a key value pair."""
    if not isinstance(var, dict):
        raise TypeError("Values can only be used on a record")
    return [v for v in var.values()]


def GetEnv(varname: str) -> str:
    """Returns the value of an environment variable"""
    return os.getenv(varname)


def SetEnv(varname: str, value: str) -> None:
    """Sets the value of an environment variable"""
    os.environ[varname] = value


def GetArgs() -> list:
    """Returns a list of arguments passed to the program"""
    return sys.argv


def GetArg(index: int) -> str:
    """Returns a single argument passed to the program"""
    return sys.argv[index]


def GetArgCount() -> int:
    """Returns the number of arguments passed to the program"""
    return len(sys.argv)




from __future__ import annotations
from typing import Optional, List

from .macal_variable import MacalVariable
from .ast_nodetype import AstNodetype

class NewScope:
    def __init__(self, name: str) -> NewScope:
        self.name = name
        self.parent: NewScope = None
        self.root: NewScope = self
        self.children = []
        self.variables = []
        self.function_aliasses = []
        self.labels = []
        self.jump_table = []
        self.switch_jump_tables = []
        self.functions: List[NewScope] = []
        self.is_function: bool = False
        self.is_extern_function_definition: bool = False
        self.function_definition: list = []
        self.libraries: List[str] = []
        self.is_library: bool = False

    def __str__(self) -> str:
        return f"{self.name}({self.parent.name if self.parent is not None else None}) {self.is_function} {self.is_function_definition}"

    def add_child(self, child: NewScope):
        self.children.append(child)
        child.parent = self
        child.root = self.root

    def new_child(self, name: str) -> NewScope:
        child = NewScope(name)
        self.add_child(child)
        return child
    
    def get_child(self, name: str) -> Optional[NewScope]:
        return next((child for child in self.children if child.name == name), None)

    def get_variable(self, name: str) -> Optional[MacalVariable]:
        var = next((var for var in self.variables if var.name == name), None)
        if var is not None:
            return var
        if not self.is_library:
            for lib in self.root.libraries:
                library = self.root.get_child(lib)
                var = library.get_variable(name)
                if var is not None:
                    return var
        if self.is_function or self.is_function:
            return None
        if self.parent is not None and not self.is_library:
            return self.parent.get_variable(name)
        return None
        
    def new_variable(self, name: str) -> MacalVariable:
        var = MacalVariable(name, self)
        var.value = AstNodetype.NIL
        self.variables.append(var)
        return var

    def discard_child(self, child: NewScope):
        self.children.remove(child)
        child.parent = None

    def get_function(self, name: str) -> Optional[NewScope]:
        func = next((func for func in self.functions if func.name == name), None)
        if func is not None:
            return func
        if not self.is_library:
            for lib in self.root.libraries:
                library = self.root.get_child(lib)
                func = library.get_function(name)
                if func is not None:
                    return func
        if self.parent is not None and not self.is_library:
            return self.parent.get_function(name)
                
    
    def new_function(self, name: str) -> NewScope:
        func = NewScope(name)
        func.is_function = True
        self.functions.append(func)
        return func




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      22-11-2023
#
# Copyright 2023 Westcon-Comstor
#

# Scope for the Macal VM

from __future__ import annotations
from typing import Optional
from .macal_variable import MacalVariable


class NewVMScope:
    def __init__(self, name: str) -> NewVMScope:
        self.name = name
        self.parent: NewVMScope = None
        self.children: list[NewVMScope] = []
        self.variables: list[MacalVariable] = []
        self.function_definitions = {}
        self.is_function: bool = False
        self.is_extern: bool = False
        self.module: str = None
        self.function: str = None
        self.root: NewVMScope = self
        self.breaking: bool = False
        self.continuing: bool = False
        self.halting: bool = False
        self.returning: bool = False
        
    def Halt(self):
        self.halting = True
        if self.parent is not None:
            self.parent.Halt()

    def print(self, indent: int = 0):
        print(" " * indent, "Name:       ", self.name)
        print(" " * indent, "Variables:  ", len(self.variables))
        print(" " * indent, "Functions:  ", len(self.function_definitions))
        print(" " * indent, "IsFunction: ", self.is_function)
        print(" " * indent, "IsExtern:   ", self.is_extern)
        print(" " * indent, "Module:     ", self.module)
        print(" " * indent, "Function:   ", self.function)
        print(" " * indent, "Breaking:   ", self.breaking)
        print(" " * indent, "Continuing: ", self.continuing)
        print(" " * indent, "Halting:    ", self.halting)
        print(" " * indent, "Returning:  ", self.returning)
        for child in self.children:
            child.print(indent + 1)

    def add_child(self, child: NewVMScope):
        self.children.append(child)
        child.parent = self
        child.root = self.root

    def new_child(self, name: str) -> NewVMScope:
        child = NewVMScope(name)
        self.add_child(child)
        return child
    
    def get_child(self, name: str) -> NewVMScope:
        return next((child for child in self.children if child.name == name), None)
    
    def discard_child(self, child: NewVMScope):
        self.children.remove(child)
        child.parent = None

    def __str__(self) -> str:
        return f"{self.name}({self.parent.name if self.parent is not None else None})"
    
    def get_variable(self, name: str) -> Optional[MacalVariable]:
        var = next((var for var in self.variables if var.name == name), None)
        if var is not None:
            return var
        if self.is_function:
            return var
        if self.parent is not None:
            return self.parent.get_variable(name)
        return None
    
    def get_function_definition(self, name: str):
        func = next((func for fname, func in self.function_definitions.items() if fname == name), None)
        if func is not None:
            return func
        if self.parent is not None and self.parent.name is self.name:
            if self.root is not self:
                return self.root.get_function_definition_from_root_up(name)
            return None
        if self.parent is not None:
            return self.parent.get_function_definition(name)
        return None
    
    def get_function_definition_from_root_up(self, name: str):
        func = next((func for fname, func in self.function_definitions.items() if fname == name), None)
        if func is not None:
            return func
        for child in self.children:
            func = child.get_function_definition_from_root_up(name)
            if func is not None:
                return func
        return None
    
    def new_variable(self, name: str) -> MacalVariable:
        var = MacalVariable(name, self)
        self.variables.append(var)
        return var



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the Macal Interpreter, it will lex, parse, compile and interpret (execute).

import os

from macal.config import SearchPath
from macal.macal_binary_file_io import BINARY_FILE_EXTENSION, SOURCE_FILE_EXTENSION, LoadBinary, SaveBinary
from macal.macal_vm import MacalVm


rt = MacalVm()
print("Hostname: ", "hostname")
print("API Key: ", "apikey")
print("Orgid: ", "orgiid")
print("Orgname: ", "hostname")
print("Script: ", "script")
print("Device serial: ", "serial")    
        
rt.SetReservedVariable('sysargv', "hostname")    
rt.SetReservedVariable('configuration', "configuration")
rt.SetReservedVariable('agent_version', "version 5.0")
rt.SetReservedVariable('org_name', "orgname")
rt.SetReservedVariable('org_id', "orgid")
rt.SetReservedVariable('api_key', "apikey")
rt.SetReservedVariable('host_name', "hostname")
rt.SetReservedVariable('serial_number', "serial")
code = LoadBinary("usereserved.mcb")
rt.Execute(code)

search_path = [
        os.path.dirname(__file__),
        os.path.join(os.path.dirname(__file__), "lib"),
        os.path.join(os.path.dirname(__file__), "lib", "ext"),
        os.getcwd(),
        os.path.join(os.getcwd(), "lib"),
        os.path.join(os.getcwd(), "lib", "ext"),
    ]
for path in search_path:
    if path not in SearchPath:
        SearchPath.append(path)





Opcode_NEW_VARIABLE = 0
Opcode_NEW_SCOPE = 1
Opcode_LOAD_CONSTANT = 2
Opcode_STORE_VARIABLE = 3
Opcode_LOAD_VARIABLE = 4
Opcode_LEAVE_SCOPE = 5
Opcode_LOAD_LIBRARY = 6
Opcode_ADD = 7
Opcode_SUB = 8
Opcode_MUL = 9 
Opcode_DIV = 10
Opcode_MOD = 11
Opcode_POW = 12
Opcode_GTE = 13
Opcode_LTE = 14
Opcode_EQ = 15
Opcode_NEQ = 16
Opcode_GT = 17
Opcode_LT = 18
Opcode_AND = 19
Opcode_OR = 20
Opcode_XOR = 21
Opcode_NOT = 22
Opcode_CALL = 23
Opcode_RET = 24
Opcode_FUNCTION_DEFINITION = 25
Opcode_EXTERN_FUNCTION_DEFINITION = 26
Opcode_HALT = 27
Opcode_PRINT = 28
Opcode_CONTINUE = 29
Opcode_BREAK = 30
Opcode_FOREACH = 31
Opcode_WHILE = 32
Opcode_IF = 33
Opcode_ELIF = 34
Opcode_ELSE = 35
Opcode_SELECT = 36
Opcode_DISTINCT = 37
Opcode_FROM = 38
Opcode_WHERE = 39
Opcode_MERGE = 40
Opcode_INTO = 41
Opcode_SWITCH = 42
Opcode_CASE = 43
Opcode_DEFAULT = 44
Opcode_INDEX = 45
Opcode_IS_TYPE = 46
Opcode_TYPE_STATEMENT = 47


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      22-11-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the Macal Virtual Machine Interpreter, it will execute the Macal code.

import importlib
import importlib.util
import os
from typing import Any, Optional, Union, List, Dict

from .ast_nodetype import AstNodetype
from .ast_node_select_field import AstNodeSelectField
from .config import SearchPath
from .macal_conversions import typeFromValue
from .macal_opcode import *
from .macal_variable import MacalVariable
from .macal_vm_scope import NewVMScope

from .python_module_info import ModuleInfo

class ExpressionResult:
    def __init__(self, value: Any, is_variable: bool, is_indexed: bool, index: Optional[List]):
        self.value = value
        self.is_variable = is_variable
        self.is_indexed = is_indexed
        self.index = index

    def __str__(self) -> str:
        return f"ExpressionResult({self.value}, {self.is_variable}, {self.is_indexed}, {self.index})"

class MacalVm:
    def __init__(self, verbose: Optional[bool] = None):
        self.scope: NewVMScope = None
        self.verbose = verbose
        self.exitcode = 0
        self.LoadedModules: Dict[str, ModuleInfo] = {}
        self.ReservedVariables: List[MacalVariable] = []

    def SetReservedVariable(self, name: str, value: Any) -> None:
        var = MacalVariable(name, None)
        if value is None:
            value = AstNodetype.NIL
        var.value = value
        var.value_type = typeFromValue(value)
        self.ReservedVariables.append(var)

    def _setup_reserved_variables(self, scope: NewVMScope) -> None:
        for var in self.ReservedVariables:
            var.scope = scope
            scope.variables.append(var)

    def Execute(self, instructions: List) -> None:
        self._execute_block(instructions, None)
    
    def _execute_block(self, block: List, scope: NewVMScope) -> int:
        # a block is a List of instructions.
        # instructions are tuples, where the first element is the opcode, and the rest are the arguments.
        # we split the instruction into opcode and arguments, and execute the opcode with the arguments.
        result = None
        if scope is not None:
            if scope.continuing:
                scope.continuing = False
            if scope.breaking:
                scope.breaking = False
            if scope.halting:
                return Opcode_HALT
        if self.verbose:
            print("execute_block: ", scope.name if scope is not None else "none")
        for instruction in block:
            op = instruction[0]
            if self.verbose:
                print(f"Block Execute (scope: {scope.name if scope is not None else 'none'}): {instruction[1:]}")
            res = self._execute_instruction(op, instruction[1:], scope)
            if op == Opcode_NEW_SCOPE or op == Opcode_LEAVE_SCOPE:
                scope = res
            if res in [Opcode_HALT, Opcode_CONTINUE, Opcode_BREAK]:
                return res
            if scope.returning:
                result = res
                break
            if scope.halting: break
            if scope.breaking: break
            if scope.continuing: break
        return result

    # Expression for value always outputs a constant value.
    def _execute_expression_for_value(self, expression, scope: NewVMScope, 
                                     expand_to_string: bool = False, 
                                     iswhere:bool = False,
                                     where_item_data: Optional[Dict] = None) -> Any:
        # Note expression needs to be an array of tuples.
        result = self._execute_expression(expression = expression,
                                         scope = scope,
                                         expand_to_string = expand_to_string,
                                         iswhere = iswhere,
                                         where_item_data = where_item_data)
        if not isinstance(result, ExpressionResult):
            raise Exception(f"Expected expression result, got {type(result)}")
        output = result.value
        if result.is_variable:
            output = result.value.value
            if result.is_indexed:
                output = self._execute_walk_index(output, result.index)
        if output == AstNodetype.NIL:
            output = "nil"
        if expand_to_string:
            output = str(output)
        return output

    # Walk index of an indexable value, output is always a constant value.
    def _execute_walk_index(self, source: Any, index_to_walk: List) -> Any:
        if source is None:
            raise ValueError("Source cannot be None")
        if not (isinstance(source, list) or isinstance(source, dict) or isinstance(source, str)):
            raise Exception(f"This variable type ({type(source)}) does not support indexing")
        if index_to_walk is None:
            raise ValueError("Index cannot be None")
        if len(index_to_walk) == 0:
            return source
        value = source
        for index in index_to_walk:
            if isinstance(value, dict):
                if not isinstance(index, str):
                    raise Exception("Record index must be a string")
                if index not in value.keys():
                    raise Exception("Index out of range")
            elif isinstance(value, list) or isinstance(value, str):
                if isinstance(index, int):
                    if index < 0 or index >= len(value):
                        raise Exception("Index out of range")
                else:
                    raise Exception("Array/string Index must be an integer")
            value = value[index]
        return value

    # Assign to index of variable
    def _execute_walk_index_for_store(self, var: MacalVariable, index_to_walk: List, value: Any) -> None:
        if not isinstance(var, MacalVariable):
            raise Exception(f"Not a variable, got: {type(var)}")
        if not (isinstance(var.value, list) or isinstance(var.value, dict) or isinstance(var.value, str)):
            raise Exception(f"This variable type ({type(var.value)}) does not support indexing")
        if index_to_walk is None:
            raise ValueError("Index cannot be None")
        if len(index_to_walk) == 0:
            var.value = value
            return
        value_to_index = var.value
        if len(index_to_walk) > 1:
            for index in index_to_walk[:-1]:
                if isinstance(value_to_index, dict):
                    if not isinstance(index, str):
                        raise Exception(f"Record index must be a string (got {type(index)})")
                    if index not in value_to_index.keys():
                        raise Exception("Index out of range")
                elif isinstance(value_to_index, list) or isinstance(value_to_index, str):
                    if isinstance(index, int):
                        if index < 0 or index >= len(value_to_index):
                            raise Exception("Index out of range")
                    else:
                        raise Exception("Array/string Index must be an integer")
                value_to_index = value_to_index[index]
        value_to_index[index_to_walk[-1]] = value       

    def _execute_expression(self, expression, 
                           scope: NewVMScope, 
                           expand_to_string: bool = False, 
                           allow_new_variable: bool = False,
                           iswhere: bool = False,
                           where_item_data: Optional[Dict] = None) -> ExpressionResult:
        # Note expression needs to be an array of tuples.
        op = expression[0][0]
        expr = expression[0][1:]
        if self.verbose:
            print("Execute expression: ", expr)
        if op == Opcode_LOAD_VARIABLE:
            new_var = False
            name = expr[0]
            var = scope.get_variable(name)
            index = None
            indexed = False
            if self.verbose:
                print()
                print(var)
                print("name: ", name)
                print("iswhere: ", iswhere)
                print("where_item_data: ", where_item_data)
                print("Where item data keys: ", where_item_data.keys() if where_item_data is not None else None)
                print("Where item data name: ", where_item_data[name] if where_item_data is not None and name in where_item_data.keys() else None)
                print()
            if var is None and iswhere and where_item_data is not None and name in where_item_data.keys():
                #print("Found where item data: ", where_item_data[name])
                return ExpressionResult(where_item_data[name], False, False, None)
            if var is None and allow_new_variable:
                var = scope.new_variable(name)
                new_var = True
            if var is None:
                raise Exception(f"Unknown variable {name}")
            if len(expr) > 1:
                if new_var is True:
                    raise Exception(f"Cannot index a new variable {name}")
                index = []
                for index_expr in expr[1]:
                    index.append(self._execute_expression_for_value([index_expr], scope, False)) 
                indexed = True
            return ExpressionResult(var, True, indexed, index)
        elif op == Opcode_LOAD_CONSTANT:
            return ExpressionResult(expr[0], False, False, None)
        elif op == Opcode_CALL:
            result = self._execute_function_call(expr, scope)
            return ExpressionResult(result, False, False, None)
        elif op == Opcode_NOT:
            lhs = self._execute_expression_for_value(expr[0], 
                                                    scope, 
                                                    iswhere= iswhere, 
                                                    where_item_data= where_item_data)
            return ExpressionResult(not lhs, False, False, None)
        elif op == Opcode_IS_TYPE:
            var = self._execute_expression(expr[0], scope)
            if var.is_variable is False:
                raise Exception("Expected variable, got constant")
            if var.value is None:
                raise Exception(f"Unknown variable {expr[0]}")
            type_to_check = expr[1]
            var_to_check: MacalVariable = var.value
            type_to_validate = typeFromValue(var_to_check.value)
            if var.is_indexed:
                type_to_validate = typeFromValue(self._execute_walk_index(var_to_check, var.index))
            return ExpressionResult(type_to_validate == type_to_check, False, False, None)
        elif op == Opcode_TYPE_STATEMENT:
            var = self._execute_expression(expr[0], scope)
            if var.is_variable is False:
                raise Exception("Expected variable, got constant")
            if var.value is None:
                raise Exception(f"Unknown variable {expr[0]}")
            var_to_check: MacalVariable = var.value
            result = typeFromValue(var_to_check.value)
            if var.is_indexed:
                result = typeFromValue(self._execute_walk_index(var_to_check, var.index))
            return ExpressionResult(result, False, False, None)
        elif len(expr) == 2:
            return self._execute_binary_expression(expression = expression,
                                                  scope = scope,
                                                  expand_to_string = expand_to_string,
                                                  iswhere = iswhere,
                                                  where_item_data = where_item_data)
        raise Exception(f"Unknown expression opcode {op}")
    
    def _execute_binary_expression(self, expression, scope: NewVMScope, 
                                  expand_to_string: bool = False,
                                  iswhere:bool = False,
                                  where_item_data: Optional[Dict] = None) -> ExpressionResult:
        op = expression[0][0]
        expr = expression[0][1:]
        if self.verbose:
            print("Execute binary expression: ", expr)
        lhs = self._execute_expression_for_value(expression = expr[0],
                                                scope = scope,
                                                expand_to_string = expand_to_string,
                                                iswhere= iswhere,
                                                where_item_data = where_item_data)
        rhs = self._execute_expression_for_value(expression = expr[1],
                                                scope = scope,
                                                expand_to_string = expand_to_string,
                                                iswhere = iswhere,
                                                where_item_data = where_item_data)
        if expand_to_string:
            lhs = str(lhs)
            rhs = str(rhs)
        if op == Opcode_ADD:
            result = lhs + rhs
        elif op == Opcode_SUB:
            result = lhs - rhs
        elif op == Opcode_MUL:
            result = lhs * rhs
        elif op == Opcode_DIV:
            result = lhs / rhs
        elif op == Opcode_MOD:
            result = lhs % rhs
        elif op == Opcode_POW:
            result = lhs ** rhs
        elif op == Opcode_GTE:
            result = lhs >= rhs
        elif op == Opcode_LTE:
            result = lhs <= rhs
        elif op == Opcode_EQ:
            result = lhs == rhs
        elif op == Opcode_NEQ:
            result = lhs != rhs
        elif op == Opcode_GT:
            result = lhs > rhs
        elif op == Opcode_LT:
            result = lhs < rhs
        elif op == Opcode_AND:
            result = lhs and rhs
        elif op == Opcode_OR:
            result = lhs or rhs
        elif op == Opcode_XOR:
            result = lhs ^ rhs
        else:
            raise Exception(f"Unknown expression opcode {op}")
        return ExpressionResult(result, False, False, None)

    def _execute_function_call(self, function, scope: NewVMScope) -> Any:
        name = function[0]
        func_args = function[2]
        func = scope.get_function_definition(name)
        if func is None:
            raise Exception(f"Unknown function {name}")
        params = func[0]
        block  = func[1]
        module = None
        ext_function = None
        if block is None:
            module = func[2]
            ext_function = func[3]
        args = []
        funcscope = scope.new_child(name)
        for i in range(len(params)):
            param = funcscope.new_variable(params[i][1])
            arg = self._execute_expression_for_value([func_args[i]], scope)
            param.value = arg
            args.append(arg)
        if len(func_args) > len(params):
            param.value = [param.value]
            for i in range(len(params), len(func_args)):
                arg = self._execute_expression_for_value([func_args[i]], scope)
                param.value.append(arg)
                args.append(arg)
        if block is not None:
            return_value = self._execute_block(block, funcscope)
        else:
            return_value = self._run_external_function(module, ext_function, args)
        scope.discard_child(funcscope)
        if self.verbose:
            print(scope.name, "returning", return_value, scope.returning)
        return return_value

    def _execute_instruction(self, op: int, instruction, scope: NewVMScope) -> NewVMScope:
        if not scope is None and scope.halting:
            return scope
        if op == Opcode_NEW_SCOPE:
            name = instruction[0]
            if scope is None:
                scope = NewVMScope(name)
                self._setup_reserved_variables(scope)
                return scope
            child = scope.new_child(name)
            return child
        elif op == Opcode_FUNCTION_DEFINITION:
            name = instruction[0]
            params = instruction[1]
            block = instruction[2]
            scope.function_definitions[name] = (params, block)      
        elif op == Opcode_EXTERN_FUNCTION_DEFINITION:
            name = instruction[0]
            params = instruction[1]
            module = instruction[2]
            function = instruction[3]
            scope.function_definitions[name] = (params, None, module, function)
        elif op == Opcode_LEAVE_SCOPE:
            if scope.parent is not None:
                return scope.parent
        elif op == Opcode_NEW_VARIABLE:
            scope.new_variable(instruction[0])
        elif op == Opcode_STORE_VARIABLE:
            self._execute_assign(instruction, scope)
        elif op == Opcode_PRINT:
            for arg in instruction[1]:
                prnt_scope = scope.new_child("print")
                value = self._execute_expression_for_value([arg], prnt_scope, True)
                scope.discard_child(prnt_scope)
                print(value, end="")
            print()
        elif op == Opcode_RET:
            scope.returning = True
            if len(instruction[0]) > 0:
                retval = self._execute_expression_for_value(instruction[0], scope)
                return retval
            return None
        elif op == Opcode_HALT:
            scope.Halt()
            if len(instruction[0]) > 0:
                self.exitcode = self._execute_expression_for_value(instruction[0], scope)
        elif op == Opcode_BREAK:
            scope.breaking = True
        elif op == Opcode_CONTINUE:
            scope.continuing = True
        elif op == Opcode_CALL:
            return self._execute_function_call(instruction, scope)
        elif op == Opcode_IF:
            if self._execute_expression_for_value(instruction[0], scope):
                return self._execute_block(instruction[1], scope)
            if len(instruction[2]) > 0:
                for elif_statement in instruction[2]:
                    if self._execute_expression_for_value(elif_statement[1], scope):
                        return self._execute_block(elif_statement[2], scope)
            if len(instruction[3]) > 0:
                return self._execute_block(instruction[3][1], scope)
        elif op == Opcode_FOREACH:
            expr = self._execute_expression_for_value(instruction[0], scope)
            if not isinstance(expr, list) and not isinstance(expr, dict) and not isinstance(expr, str):
                raise Exception(f"Expected array, record or string, got {type(expr)}")
            foreach_scope = scope.new_child("foreach")
            it = foreach_scope.new_variable("it")
            for item in expr:
                it.value = item
                result = self._execute_block(instruction[1], foreach_scope)
                if result == Opcode_BREAK or result == Opcode_HALT:
                    break
                if result == Opcode_CONTINUE:
                    continue
            scope.discard_child(foreach_scope)
        elif op == Opcode_WHILE:
            while_scope = scope.new_child("while")
            while self._execute_expression_for_value(instruction[0], scope):
                result = self._execute_block(instruction[1], while_scope)
                if while_scope.breaking or while_scope.halting or while_scope.returning:
                    break
                if while_scope.continuing:
                    continue
            if while_scope.halting:
                scope.halting = True
            scope.discard_child(while_scope)
        elif op == Opcode_SWITCH:
            expr = self._execute_expression_for_value(instruction[0], scope)
            if expr in instruction[1].keys():
                return self._execute_block(instruction[1][expr], scope)
            if len(instruction[2]) > 0:
                return self._execute_block(instruction[2], scope)
        elif op == Opcode_SELECT:
            return self._execute_select(instruction, scope)
        else:
            raise Exception("Unknown instruction: ", op, " @ ", scope.name if scope is not None else "none", " ", instruction)
        return scope

    def _execute_assign(self, instruction, scope: NewVMScope):
        var_instruction    = instruction[0]
        rhs_instruction    = instruction[1]
        append = True if len(instruction) > 2 and instruction[2] else False
        if self.verbose:
            print()
            print("Setting up the assignment:")
            print()
            print("var_instruction    = ", var_instruction)
            print("var name:          = ", var_instruction[0][1])
            print("Index:             = ", var_instruction[0][2:] if len(var_instruction[0]) > 2 else '')
            print("rhs_instruction    = ", rhs_instruction)        
            print("append             = ", append)
            print()
        result = self._execute_expression(var_instruction, scope)
        if result.is_variable is False:
            raise Exception("Expected variable, got constant")
        if result.value is None:
            raise Exception(f"Unknown variable {var_instruction[0][1]}")
        variable_to_assign_to: MacalVariable = result.value
        if self.verbose:
            print("variable_to_assign_to = ", variable_to_assign_to)
            print("Result is indexed:     ", result.is_indexed)
            print("Result index:          ", result.index)
        value_to_assign = self._execute_expression_for_value(rhs_instruction, scope)
        if self.verbose:
            print("performed the calcualtions:")
            if not result.is_indexed:
                print(f"var ({variable_to_assign_to.name}) = {value_to_assign}")
            else:
                print(f"{variable_to_assign_to.name}{result.index} = {value_to_assign}")
            print("value to be assigned = ", value_to_assign)
        if result.is_indexed:
            self._execute_walk_index_for_store(variable_to_assign_to, result.index, value_to_assign)
        else:
            if append:
                variable_to_assign_to.value.append(value_to_assign)
            else:
                variable_to_assign_to.value = value_to_assign

    def _execute_select(self, instruction, scope: NewVMScope):
        fields: List[AstNodeSelectField] = instruction[0]
        where_expr = instruction[2]
        distinct = instruction[3]
        merge = instruction[4]
        destination_data = []

        # get the destination data if we do a merge:
        if merge:
            destination_data = self._execute_expression_for_value(instruction[5], scope)
        # gets the destination variable, allow creation of a new variable if we do not merge.
        into_expr_result = self._execute_expression(instruction[5], scope, allow_new_variable=not merge)
        if into_expr_result.is_variable is False:
            raise Exception("Select into expected variable, got constant")
        into_var: MacalVariable = into_expr_result.value
        # we get the source data by executing the from instruction.
        source_data = self._execute_expression_for_value(instruction[1], scope)
        if source_data is None or source_data is False or source_data == AstNodetype.NIL or source_data == "nil":
            source_data = []
        # check if we have an array or record, if not, we throw an exception.
        if not isinstance(source_data, list) and not isinstance(source_data, dict):
            raise Exception(f"From expected array or record, got {type(source_data)}")
        if not isinstance(destination_data, List) and not isinstance(destination_data, dict):
            raise Exception(f"Expected array, record or string, got {type(destination_data)}")

        # convert the source data to an array if it is a record.
        if isinstance(source_data, dict):
            source_data = [source_data]

        # if we have a where expression, we filter the source data.
        result_data = []

        if len(where_expr) > 0:
            for item in source_data:
                if self._execute_expression_for_value(where_expr, scope, iswhere=True, where_item_data=item) is False:
                    continue
                if distinct and item not in result_data:
                    result_data.append(item)
                elif not distinct:
                    result_data.append(item)
        else:
            result_data = source_data.copy()
        # source data is filtered into the result data.
        # we will now apply the field filters to the result data if we do not have a select all (*).
        if not (len(fields) == 1 and fields[0].fieldname == "*"):
            result_data = self._apply_field_filters(fields, result_data)

        # distinct always results in a single record no matter what.
        # also if there is only 1 record in the result data, we also return it as a single record.
        if isinstance(result_data, list) and distinct is True and len(result_data) > 0:
            result_data = result_data[0]

        # if we have a merge, we merge the result data into the destination data.
        if merge is True:
            destination_data = self._merge_data(destination_data, result_data).copy()
        else:
            destination_data = result_data.copy()

        # distinct always results in a single record no matter what.
        # also if there is only 1 record in the result data, we also return it as a single record.
        if isinstance(destination_data, list) and distinct is True and len(destination_data) > 0:
            destination_data = destination_data[0]

        if distinct is True and len(destination_data) == 1:
            destination_data = destination_data[list(destination_data.keys())[0]]

        # we now have the final result data, we store it into the destination variable.
        # we need to take into account if the destination variable has an index or not.
        if into_expr_result.is_indexed:
            self._execute_walk_index_for_store(into_var, into_expr_result.index, destination_data)
        else:
            into_var.value = destination_data

    def _apply_field_filters(self, fields: List[AstNodeSelectField], source_data: List[Dict]) -> List[Dict]:
        if len(source_data) == 0:
            return [{field.altfieldname: AstNodetype.NIL for field in fields}]
        return [{field.altfieldname: item.get(field.fieldname, AstNodetype.NIL) for field in fields} for item in source_data]

    def _merge_data(self, source_data: List[Dict], destination_data: List[Dict]) -> Union[List[Dict], Dict]:
        if isinstance(source_data, dict):
            source_data = [source_data]
        if isinstance(destination_data, dict):
            destination_data = [destination_data]
        if not (isinstance(source_data, list) and isinstance(destination_data, list)):
            raise Exception("MERGE: Type error, both parameters must be arrays.")
        if len(source_data)== 0 and len(destination_data) == 0:
            return []
        if len(source_data) == 0 and len(destination_data) > 0:
            final_data = destination_data
        elif len(source_data) > 0 and len(destination_data) == 0:
            final_data = source_data
        elif len(source_data) > 1 and len(destination_data) > 1 and set(source_data[0].keys()) == set(destination_data[0].keys()):
            final_data = [source_data[0], destination_data[0]]
        elif len(source_data) == 1 and len(destination_data) == 1 and set(source_data[0].keys()) != set(destination_data[0].keys()):
            keys = set().union(source_data[0].keys(), destination_data[0].keys())
            final_data = [{k: source_data[0].get(k, destination_data[0].get(k, AstNodetype.NIL)) for k in keys}]
        # multiple records in each, with the same set of fields, then just append them both.
        elif set(source_data[0].keys()) == set(destination_data[0].keys()):
            final_data = source_data.copy()
            final_data.extend(destination_data)
        # multiple records in each, but with the or different fields, then merge the records and fields.
        elif set(source_data[0].keys()) != set(destination_data[0].keys()):
            keys = set().union(*(d.keys() for d in source_data + destination_data))
            final_data = [{k: d.get(k, AstNodetype.NIL) for k in keys} for d in source_data + destination_data]
        if len(final_data) == 1:
            return final_data[0]
        return final_data

    def _import_module(self, module_name: str) -> Optional[Any]:
        try:
            return importlib.import_module(module_name)
        except ImportError:
            return None

    def _import_module_from_path(self, module_name: str) -> Optional[Any]:
        try:
            for path in SearchPath:
                path = os.path.join(path, f"{module_name}.py")
                if not os.path.exists(path): continue
                spec = importlib.util.spec_from_file_location(module_name, path)
                if spec is None: continue
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                return module
        except ImportError as ex:
            self._Error(f"Import error: {ex}")
        return None

    def _run_external_function(self, module_name: str, function_name: str, args ) -> Any:
        module = self.LoadedModules.get(module_name, None)
        if module is None:
            imported_module = self._import_module(module_name)
            if imported_module is None:
                imported_module = self._import_module_from_path(module_name)
                if imported_module is None:
                    raise Exception(f"Module {module_name} not found.")
            module = ModuleInfo(module_name, imported_module)
            self.LoadedModules[module_name] = module
        function = module.functions.get(function_name, None)
        if function is None:
            raise Exception(f"Function {function_name} not found in module {module_name}.")
        result = function(*args)
        if result is None:
            result = AstNodetype.NIL
        return result
        



# tests all mcl files in the tests directory

import os
import sys
import subprocess
from keyring import get_password
from macal.config import SearchPath
from macal.macal_vm import MacalVm
from macal.macal_binary_file_io import BINARY_FILE_EXTENSION, SOURCE_FILE_EXTENSION, LoadBinary

def setSearchPath():
    search_path = [
        os.path.dirname(__file__),
        os.path.join(os.path.dirname(__file__), "lib"),
        os.path.join(os.path.dirname(__file__), "lib", "ext"),
        os.getcwd(),
        os.path.join(os.getcwd(), "lib"),
        os.path.join(os.getcwd(), "lib", "ext"),
    ]
    for path in search_path:
        if path not in SearchPath:
            SearchPath.append(path)

def execute(code):
    api_key = get_password("meraki", "Stage Entertainment")
    rt = MacalVm()
    rt.SetReservedVariable('sysargv', "hostname")    
    rt.SetReservedVariable('configuration', "configuration")
    rt.SetReservedVariable('agent_version', "version 5.0")
    rt.SetReservedVariable('org_name', "Stage Entertainment")
    rt.SetReservedVariable('org_id', None)
    rt.SetReservedVariable('api_key', api_key)
    rt.SetReservedVariable('host_name', "Stage-Entertainment-Master")
    rt.SetReservedVariable('serial_number', "serial")
    rt.Execute(code)
    return rt.exitcode

def main():
    # get all files in the tests directory
    file_list = os.listdir('.')

    # remove non-mcl files
    source_files = [test for test in file_list if test.endswith(SOURCE_FILE_EXTENSION)]
    # sort the tests
    source_files.sort()

    # run each test
    print('Compiling tests: ')
    good = 0
    errors = 0
    for test in source_files:
        if test in ['test4.mcl','function.mcl','function6.mcl']:
            continue
        # get the test name
        test_name = test[:-4]
        # run the test
        p = subprocess.run(['mc', test, '-r', 'sysargv,configuration,agent_version,org_name,org_id,api_key,host_name,serial_number,script'],capture_output=True)
        if p.returncode != 0:
            print(f'Error compiling {test_name}:')
            print(p.stderr.decode('utf-8'))
            errors += 1
        else:
            good += 1
    print(f'Good: {good} Errors: {errors}, Total: {good+errors}')
    
    file_list = os.listdir('.')
    executables = [test for test in file_list if test.endswith(BINARY_FILE_EXTENSION)]
    executables.sort()

    print('Executing tests: ')
    good = 0
    errors = 0
    errors_list = []
    halt_count = 0
    halted = []
    for test in executables:
        test_name = test[:-4]
        code = LoadBinary(test)
        try:
            ret = execute(code)
            if ret != 0:
                halted.append((test_name, ret))
                halt_count += 1
            good += 1
        except Exception as e:
            errors_list.append((test_name, e))
            errors += 1
    
    print()
    print(f'Good: {good} Errors: {errors}, halted: {halt_count}, Total: {good+errors}')
    print()

    for error in errors_list:
        print(f'Error executing {error[0]}:')
        print(error[1])

    print()
    print('Tests that exit with halt:')
    for test in halted:
        print(f"Test: '{test[0]}', exitcode: {test[1]}")


if __name__ == '__main__':
    main()


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the new Macal Compiler, it will compile the AST from the parser into "bytecode" for the VM.

from __future__ import annotations

import os
from typing import Optional, List

from .ast_node_assignment import AstNodeAssignmentStatement
from .ast_node_binary_expression import AstNodeBinaryExpression
from .ast_node_block import AstNodeBlock
from .ast_node_break_statement import AstNodeBreakStatement
from .ast_node_case_statement import AstNodeCaseStatement
from .ast_node_continue_statement import AstNodeContinueStatement
from .ast_node_elif_statement import AstNodeElifStatement
from .ast_node_expression import AstNodeExpression
from .ast_node_foreach_statement import AstNodeForeachStatement
from .ast_node_function_call_statement import AstNodeFunctionCallStatement
from .ast_node_function_call_expression import AstNodeFunctionCallExpression
from .ast_node_function_definition import AstNodeFunctionDefinition
from .ast_node_halt_statement import AstNodeHaltStatement
from .ast_node_if_statement import AstNodeIfStatement
from .ast_node_istype import AstNodeIsType
from .ast_node_include_statement import AstNodeIncludeStatement
from .ast_node_indexed_variable_expression import AstNodeIndexedVariableExpression
from .ast_node_literal_expression import AstNodeLiteralExpression
from .ast_node_print_statement import AstNodePrintStatement
from .ast_node_program import AstNodeProgram
from .ast_node_return_statement import AstNodeReturnStatement
from .ast_node_select_statement import AstNodeSelectStatement
from .ast_node_statement import AstNodeStatement
from .ast_node_switch_statement import AstNodeSwitchStatement
from .ast_node_type_statement import AstNodeTypeStatement
from .ast_node_unary_expression import AstNodeUnaryExpression
from .ast_node_variable_expression import AstNodeVariableExpression
from .ast_node_variable_function_call_expression import AstNodeVariableFunctionCallExpression
from .ast_node_while_statement import AstNodeWhileStatement
from .ast_nodetype import AstNodetype
from .config import SearchPath
from .macal_compiler_scope import NewScope
from .macal_lexer import Lexer
from .macal_opcode import *
from .macal_parser import Parser
from .macal_variable import MacalVariable

class MacalCompiler:
    def __init__(self, verbose: bool):
        self.program: Optional[AstNodeProgram] = None
        self.output: str = None
        self.verbose: bool = verbose
        self.reserved: list[str] = []

    def compile(self, program: AstNodeProgram, reserved: list[str]) -> list:
        self.program = program
        self.reserved = reserved
        scope = NewScope("root")
        instructions = [(Opcode_NEW_SCOPE, "root")]
        for var in reserved:
            scope.new_variable(var.strip())
            instructions.append((Opcode_NEW_VARIABLE, var.strip()))
        self.compile_program(self.program, scope, instructions)
        instructions.append((Opcode_LEAVE_SCOPE, "root"))
        instructions.append((Opcode_HALT, [(Opcode_LOAD_CONSTANT, 0)])) # not really needed, but for clarity and completeness.
        if self.verbose:
            for instruction in instructions:
                print(instruction[0], " ", instruction[1:])
        return instructions

    def compile_program(self, program: AstNodeProgram, scope: NewScope, instructions: list):
        for statement in program.statements:
            self.compile_statement(statement, scope, instructions)

    def compile_statement(self, statement: AstNodeStatement, scope: NewScope, instructions: list):
        if statement.expr_type == AstNodetype.ASSIGNMENT_STATEMENT:
            self.compile_assignment_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.PRINT_STATEMENT:
            self.compile_print_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.FUNCTION_DEFINITION:
            self.compile_function_definition(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.RETURN_STATEMENT:
            self.compile_return_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.FUNCTION_CALL:
            self.compile_function_call_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.FUNCTION_CALL_EXPRESSION:
            self.compile_function_call_expression(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.HALT_STATEMENT:
            self.compile_halt_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.BREAK_STATEMENT:
            self.compile_break_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.CONTINUE_STATEMENT:
            self.compile_continue_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.IF_STATEMENT:
            self.compile_if_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.WHILE_STATEMENT:
            self.compile_while_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.FOREACH_STATEMENT:
            self.compile_foreach_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.SWITCH_STATEMENT:
            self.compile_switch_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.INCLUDE_STATEMENT:
            self.compile_include_statement(statement, scope, instructions)
        elif statement.expr_type == AstNodetype.SELECT_STATEMENT:
            self.compile_select_statement(statement, scope, instructions)
        else:
            raise Exception(f"Unknown statement type {statement.expr_type} @ {statement.line}:{statement.column}")

    def compile_block(self, statement: AstNodeBlock, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.BLOCK:
            raise Exception(f"Expected block, got {statement.expr_type} @ {statement.line}:{statement.column}")
        #child = scope.new_child("block")
        #instructions.append((Opcode_NEW_SCOPE, child.name))
        for statement in statement.statements:
            self.compile_statement(statement, scope, instructions)
        #instructions.append((Opcode_LEAVE_SCOPE, child.name))

    # statement fucntions

    def compile_print_statement(self, statement: AstNodePrintStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.PRINT_STATEMENT:
            raise Exception(f"Expected print statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        args = []
        for arg in statement.args:
            self.compile_expression(arg, scope, args)
        instructions.append((Opcode_PRINT, len(statement.args), args))

    def compile_function_definition(self, statement: AstNodeFunctionDefinition, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.FUNCTION_DEFINITION:
            raise Exception(f"Expected function definition, got {statement.expr_type} @ {statement.line}:{statement.column}")
        child = scope.new_child(statement.name)
        scope.functions.append(child)
        child.is_function = True
        params = []
        for param in statement.parameters:
            params.append((Opcode_NEW_VARIABLE, param.name))
            child.variables.append(MacalVariable(param.name, child))

        fndef = None
        if statement.body is not None:
            blk = []
            self.compile_block(statement.body, child, blk)
            fndef = (Opcode_FUNCTION_DEFINITION, child.name, params, blk)

        if statement.external:
            fndef = (Opcode_EXTERN_FUNCTION_DEFINITION, statement.name, params, statement.module, statement.function)
            child.is_extern_function_definition = True

        if fndef is None:
            raise Exception(f"Function definition without body @ {statement.line}:{statement.column}")
        
        child.function_definition.append(fndef)
        instructions.append(fndef)
       
    def compile_assignment_statement(self, statement: AstNodeAssignmentStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.ASSIGNMENT_STATEMENT:
            raise Exception(f"Expected assignment statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        var = scope.get_variable(statement.lhs.name)
        if var is None:
            instructions.append((Opcode_NEW_VARIABLE, statement.lhs.name))
            scope.variables.append(MacalVariable(statement.lhs.name, scope))
        lhs = []
        self.compile_expression(statement.lhs, scope, lhs)
        op = statement.operator
        rhs = []
        self.compile_expression(statement.rhs, scope, rhs)
        if op == "=":
            instructions.append((Opcode_STORE_VARIABLE, lhs, rhs, statement.append))
            return
        if statement.append:
            raise Exception(f"Operator {op} not supported for append to array @ {statement.line}:{statement.column}")
        opcode = 0
        if op == "+=":
            opcode = Opcode_ADD
        elif op == "-=":
            opcode = Opcode_SUB
        elif op == "*=":
            opcode = Opcode_MUL
        elif op == "/=":
            opcode = Opcode_DIV
        elif op == "%=":
            opcode = Opcode_MOD
        elif op == "^=":
            opcode = Opcode_POW
        else:
            raise Exception(f"Unknown assignment operator {op} @ {statement.line}:{statement.column}")
        instructions.append((Opcode_STORE_VARIABLE, lhs, [(opcode, lhs, rhs)]))

    def compile_return_statement(self, statement: AstNodeReturnStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.RETURN_STATEMENT:
            raise Exception(f"Expected return statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        value = []
        if statement.expr is not None:
            self.compile_expression(statement.expr, scope, value)
        instructions.append((Opcode_RET, value))

    def compile_halt_statement(self, statement: AstNodeHaltStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.HALT_STATEMENT:
            raise Exception(f"Expected halt statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        value = []
        if statement.expr is not None:
            self.compile_expression(statement.expr, scope, value)
        instructions.append((Opcode_HALT, value))

    def compile_break_statement(self, statement: AstNodeBreakStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.BREAK_STATEMENT:
            raise Exception(f"Expected break statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        instructions.append((Opcode_BREAK,))

    def compile_continue_statement(self, statement: AstNodeContinueStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.CONTINUE_STATEMENT:
            raise Exception(f"Expected continue statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        instructions.append((Opcode_CONTINUE,))

    def compile_if_statement(self, statement: AstNodeIfStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.IF_STATEMENT:
            raise Exception(f"Expected if statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        expr = []
        self.compile_expression(statement.condition, scope, expr)
        blk = []
        self.compile_block(statement.block, scope, blk)
        elifs = []
        for elif_statement in statement.elif_block_list:
            self.compile_elif_statement(elif_statement, scope, elifs)
        elseblk = []
        if statement.else_block is not None:
            block = statement.else_block
            block.expr_type = AstNodetype.BLOCK
            self.compile_block(statement.else_block, scope, elseblk)
        instructions.append((Opcode_IF, expr, blk, elifs, (Opcode_ELSE, elseblk)))

    def compile_elif_statement(self, statement: AstNodeElifStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.ELIF_STATEMENT:
            raise Exception(f"Expected elif statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        expr = []
        self.compile_expression(statement.condition, scope, expr)
        blk = []
        self.compile_block(statement.block, scope, blk)
        instructions.append((Opcode_ELIF, expr, blk))

    def compile_foreach_statement(self, statement: AstNodeForeachStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.FOREACH_STATEMENT:
            raise Exception(f"Expected foreach statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        expr = []
        self.compile_expression(statement.expr, scope, expr)
        blk = []
        fescope = scope.new_child("foreach")
        fescope.new_variable("it")
        self.compile_block(statement.block, fescope, blk)
        scope.discard_child(fescope)
        instructions.append((Opcode_FOREACH, expr, blk))

    def compile_while_statement(self, statement: AstNodeWhileStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.WHILE_STATEMENT:
            raise Exception(f"Expected while statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        expr = []
        self.compile_expression(statement.condition, scope, expr)
        blk = []
        self.compile_block(statement.block, scope, blk)
        instructions.append((Opcode_WHILE, expr, blk))

    def compile_switch_statement(self, statement: AstNodeSwitchStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.SWITCH_STATEMENT:
            raise Exception(f"Expected select statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        expr = []
        self.compile_expression(statement.expr, scope, expr)
        
        switch = {}
        for case in statement.cases:
            csw = []
            self.compile_case_statement(case, scope, csw)
            if csw[0][1] in switch.keys():
                raise Exception(f"Duplicate case statement {csw[0][1]} @ {statement.line}:{statement.column}")
            switch[csw[0][1]] = csw[0][2]
        default = []
        if statement.default is not None:
            self.compile_block(statement.default.block, scope, default)
        instructions.append((Opcode_SWITCH, expr, switch, default))

    def compile_case_statement(self, statement: AstNodeCaseStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.CASE_STATEMENT:
            raise Exception(f"Expected case statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        expr = []
        self.compile_expression(statement.expr, scope, expr)
        if expr[0][0] != Opcode_LOAD_CONSTANT:
            raise Exception(f"Case expression must be a constant, got {expr[0][0]} @ {statement.line}:{statement.column}")
        blk = []
        self.compile_block(statement.block, scope, blk)
        instructions.append((Opcode_CASE, expr[0][1], blk))

    def compile_select_statement(self, statement: AstNodeSelectStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.SELECT_STATEMENT:
            raise Exception(f"Expected select statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        from_expr = []    
        self.compile_expression(statement.From, scope, from_expr)
        where_expr = []
        if statement.Where is not None:
            self.compile_expression(statement.Where, scope, where_expr, True)
        into_expr = []
        self.compile_expression(statement.Into, scope, into_expr, True)
        instructions.append((Opcode_SELECT, statement.Fields, from_expr, where_expr, statement.distinct, statement.merge, into_expr))

    def compile_function_call_statement(self, statement: AstNodeFunctionCallStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.FUNCTION_CALL:
            raise Exception(f"Expected function call, got {statement.expr_type} @ {statement.line}:{statement.column}")
        func = scope.get_function(statement.name)
        if func is None:
            raise Exception(f"Unknown function {statement.name} @ {statement.line}:{statement.column}")
        args = []
        for arg in statement.args:
            self.compile_expression(arg, scope, args)
        instructions.append((Opcode_CALL, statement.name, len(statement.args), args))

    def compile_is_type_statement(self, statement: AstNodeIsType, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.IS_TYPE_STATEMENT:
            raise Exception(f"Expected is type statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        expr = []
        self.compile_expression(statement.expr, scope, expr)
        instructions.append((Opcode_IS_TYPE, expr, statement.TypeToCheck))

    def compile_type_statement(self, statement: AstNodeTypeStatement, scope: NewScope, instructions: list):
        if statement.expr_type != AstNodetype.TYPE_STATEMENT:
            raise Exception(f"Expected type statement, got {statement.expr_type} @ {statement.line}:{statement.column}")
        expr = []
        self.compile_expression(statement.expr, scope, expr)
        instructions.append((Opcode_TYPE_STATEMENT, expr))

    # Include implementation

    def find_library(self, name: str) -> Optional[str]:
        for path in SearchPath:
            lib_path = os.path.join(path, f"{name}.mcl")
            if os.path.exists(lib_path):
                return lib_path
        return None

    def load_library(self, path: str) -> Optional[str]:
        if not os.path.exists(path):
            return None
        with open(path, "r") as f:
            return f.read()

    def compile_include_statement(self, statement: AstNodeIncludeStatement, scope: NewScope, instructions: list) -> None:
        for lib in statement.libraries:
            if lib.name in scope.libraries:
                continue
            lib_path = self.find_library(lib.name)
            if lib_path is None:
                raise Exception(f"Library {lib.name} not found at {lib.line}:{lib.column}")
            source = self.load_library(lib_path)
            if source is None:
                raise Exception(f"Library {lib.name} not found at {lib.line}:{lib.column}")
            
            scope.libraries.append(lib.name)
            lib_scope = scope.new_child(lib.name)
            lib_scope.is_library = True
            instructions.append((Opcode_NEW_SCOPE, lib.name))
            lex = Lexer(source)
            tokens = lex.tokenize()
            parser = Parser(tokens, lib_path)
            ast = parser.parse()
            self.compile_program(ast, lib_scope, instructions)

    # expression functions.

    def compile_function_call_expression(self, expression: AstNodeFunctionCallExpression, scope: NewScope, instructions: list):
        if expression.expr_type != AstNodetype.FUNCTION_CALL_EXPRESSION:
            raise Exception(f"Expected function call expression, got {expression.expr_type} @ {expression.line}:{expression.column}")
        args = []
        func = scope.get_function(expression.name)
        if func is None:
            raise Exception(f"Unknown function {expression.name} @ {expression.line}:{expression.column}")
        for arg in expression.args:
            self.compile_expression(arg, scope, args)
        instructions.append((Opcode_CALL, expression.name, len(expression.args), args))

    def compile_expression(self, expression: AstNodeExpression, scope: NewScope, instructions: list, allow_new_variable: bool = False):
        if expression.expr_type == AstNodetype.BINARY_EXPRESSION:
            self.compile_binary_expression(expression, scope, instructions, allow_new_variable=allow_new_variable)
        elif expression.expr_type == AstNodetype.LITERAL_EXPRESSION:
            self.compile_literal_expression(expression, scope, instructions)
        elif expression.expr_type == AstNodetype.UNARY_EXPRESSION:
            self.compile_unary_expression(expression, scope, instructions)
        elif expression.expr_type == AstNodetype.VARIABLE_EXPRESSION:
            self.compile_variable_expression(expression, scope, instructions, allow_new_variable=allow_new_variable)
        elif expression.expr_type == AstNodetype.FUNCTION_CALL_EXPRESSION:
            self.compile_function_call_expression(expression, scope, instructions)
        elif expression.expr_type == AstNodetype.INDEXED_VARIABLE_EXPRESSION:
            self.compile_indexed_variable_expression(expression, scope, instructions)
        elif expression.expr_type == AstNodetype.VARIABLE_FUNCTION_CALL_EXPRESSION:
            self.compile_variable_function_call_expression(expression, scope, instructions)
        elif expression.expr_type == AstNodetype.IS_TYPE_STATEMENT:
            self.compile_is_type_statement(expression, scope, instructions)
        elif expression.expr_type == AstNodetype.TYPE_STATEMENT:
            self.compile_type_statement(expression, scope, instructions)
        else:
            raise Exception(f"Unknown expression type {expression.expr_type} @ {expression.line}:{expression.column}")

    def compile_binary_expression(self, expression: AstNodeBinaryExpression, scope: NewScope, instructions: list, allow_new_variable: bool = False):
        if expression.expr_type != AstNodetype.BINARY_EXPRESSION:
            raise Exception(f"Expected binary expression, got {expression.expr_type} @ {expression.line}:{expression.column}")
        lhs = []
        self.compile_expression(expression.left, scope, lhs, allow_new_variable=allow_new_variable)
        rhs = []
        self.compile_expression(expression.right, scope, rhs, allow_new_variable=allow_new_variable)
        op = expression.operator
        if op == "+=" or op == "+":
            instructions.append((Opcode_ADD, lhs, rhs))
        elif op == "-=" or op == "-":
            instructions.append((Opcode_SUB, lhs, rhs))
        elif op == "*=" or op == "*":
            instructions.append((Opcode_MUL, lhs, rhs))
        elif op == "/=" or op == "/":
            instructions.append((Opcode_DIV, lhs, rhs))
        elif op == "%=" or op == "%":
            instructions.append((Opcode_MOD, lhs, rhs))
        elif op == "^=" or op == "^":
            instructions.append((Opcode_POW, lhs, rhs))
        elif op == ">=":
            instructions.append((Opcode_GTE, lhs, rhs))
        elif op == "<=":
            instructions.append((Opcode_LTE, lhs, rhs))
        elif op == "==":
            instructions.append((Opcode_EQ, lhs, rhs))
        elif op == "!=":
            instructions.append((Opcode_NEQ, lhs, rhs))
        elif op == ">":
            instructions.append((Opcode_GT, lhs, rhs))
        elif op == "<":
            instructions.append((Opcode_LT, lhs, rhs))
        elif op == "and":
            instructions.append((Opcode_AND, lhs, rhs))
        elif op == "or":
            instructions.append((Opcode_OR, lhs, rhs))
        elif op == "xor":
            instructions.append((Opcode_XOR, lhs, rhs))
        elif op == "not":
            instructions.append((Opcode_NOT, lhs, rhs))
        elif op == "&&":
            instructions.append((Opcode_AND, lhs, rhs))
        elif op == "||":
            instructions.append((Opcode_OR, lhs, rhs))
        elif op == "^^":
            instructions.append((Opcode_XOR, lhs, rhs))
        else:
            raise Exception(f"Unknown binary operator {op} @ {expression.line}:{expression.column}")

    def compile_literal_expression(self, expression: AstNodeLiteralExpression, scope: NewScope, instructions: list):
        if expression.expr_type != AstNodetype.LITERAL_EXPRESSION:
            raise Exception(f"Expected literal expression, got {expression.expr_type} @ {expression.line}:{expression.column}")
        instructions.append((Opcode_LOAD_CONSTANT, expression.value))

    def compile_variable_expression(self, expression: AstNodeVariableExpression, scope: NewScope, instructions: list, allow_new_variable: bool = False):
        if expression.expr_type != AstNodetype.VARIABLE_EXPRESSION:
            raise Exception(f"Expected variable expression, got {expression.expr_type} @ {expression.line}:{expression.column}")
        var = scope.get_variable(expression.name)
        if var is None and allow_new_variable:
            var = scope.new_variable(expression.name)
        if var is None:
            raise Exception(f"Unknown variable {expression.name} @ {expression.line}:{expression.column}")
        else:
            instructions.append((Opcode_LOAD_VARIABLE, expression.name))

    def compile_indexed_variable_expression(self, expression: AstNodeIndexedVariableExpression, scope: NewScope, instructions: list):
        if expression.expr_type != AstNodetype.INDEXED_VARIABLE_EXPRESSION:
            raise Exception(f"Expected indexed variable expression, got {expression.expr_type} @ {expression.line}:{expression.column}")
        var = scope.get_variable(expression.name)
        if var is None:
            raise Exception(f"Unknown variable {expression.name} @ {expression.line}:{expression.column}")
        var_index = []
        for index in expression.index:
            self.compile_expression(index, scope, var_index)
        instructions.append((Opcode_LOAD_VARIABLE, expression.name, var_index))        

    def compile_variable_function_call_expression(self, expression: AstNodeVariableFunctionCallExpression, scope: NewScope, instructions: list):
        print("compile_variable_function_call_expression")
        if expression.expr_type != AstNodetype.VARIABLE_FUNCTION_CALL_EXPRESSION:
            raise Exception(f"Expected variable function call expression, got {expression.expr_type} @ {expression.line}:{expression.column}")
        print(expression.name)
        for arg in expression.args:
            self.compile_expression(arg, scope, instructions)
        raise Exception("variable function call expression Not implemented")

    def compile_unary_expression(self, expression: AstNodeUnaryExpression, scope: NewScope, instructions: list):
        if self.verbose:
            print("compile_unary_expression")
        if expression.expr_type != AstNodetype.UNARY_EXPRESSION:
            raise Exception(f"Expected unary expression, got {expression.expr_type} @ {expression.line}:{expression.column}")
        if self.verbose:
            print(expression.operator)
        op = expression.operator
        rhs = []
        self.compile_expression(expression.expr, scope, rhs)
        if op == "-":
            instructions.append((Opcode_SUB, rhs))
        elif op == "+":
            instructions.append((Opcode_ADD, rhs))
        elif op == "not":
            instructions.append((Opcode_NOT, rhs))
        else:
            raise Exception(f"Unknown unary operator {op} @ {expression.line}:{expression.column}")




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-11-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the Macal decompiler, it will decompile a .mbc file into bytecode mnenomics.

from typing import Any
import sys

from .macal_opcode import *
from .macal_binary_file_io import BINARY_FILE_EXTENSION, LoadBinary


opcode_translation_table = {
Opcode_NEW_VARIABLE : "NEW_VARIABLE",
Opcode_NEW_SCOPE : "NEW_SCOPE",
Opcode_LOAD_CONSTANT : "LOAD_CONSTANT",
Opcode_STORE_VARIABLE : "STORE_VARIABLE",
Opcode_LOAD_VARIABLE : "LOAD_VARIABLE",
Opcode_LEAVE_SCOPE : "LEAVE_SCOPE",
Opcode_LOAD_LIBRARY : "LOAD_LIBRARY",
Opcode_ADD : "ADD",
Opcode_SUB : "SUB",
Opcode_MUL : "MUL",
Opcode_DIV : "DIV",
Opcode_MOD : "MOD",
Opcode_POW : "POW",   
Opcode_GTE : "GTE",
Opcode_LTE : "LTE",
Opcode_EQ : "EQ",
Opcode_NEQ : "NEQ",
Opcode_GT : "GT",
Opcode_LT : "LT",
Opcode_AND : "AND",
Opcode_OR : "OR",
Opcode_XOR : "XOR",
Opcode_NOT : "NOT",
Opcode_CALL : "CALL",
Opcode_RET : "RET",
Opcode_FUNCTION_DEFINITION : "FUNCTION_DEFINITION",
Opcode_EXTERN_FUNCTION_DEFINITION : "EXTERN_FUNCTION_DEFINITION",
Opcode_HALT : "HALT",
Opcode_PRINT : "PRINT",
Opcode_CONTINUE : "CONTINUE",
Opcode_BREAK : "BREAK",
Opcode_FOREACH : "FOREACH",
Opcode_WHILE :  "WHILE",
Opcode_IF : "IF",
Opcode_ELIF : "ELIF",
Opcode_ELSE : "ELSE",
Opcode_SELECT : "SELECT",
Opcode_DISTINCT : "DISTINCT",
Opcode_FROM : "FROM",
Opcode_WHERE :  "WHERE",
Opcode_MERGE : "MERGE",
Opcode_INTO : "INTO",
Opcode_SWITCH : "SWITCH",
Opcode_CASE : "CASE",
Opcode_DEFAULT : "DEFAULT",
Opcode_INDEX : "INDEX",
Opcode_IS_TYPE : "IS_TYPE",
Opcode_TYPE_STATEMENT : "TYPE_STATEMENT",
}

class MacalDecompiler:
    def __init__(self, verbose: bool = False):
        self.verbose = verbose

    def Decompile(self, code: list) -> None:
        for instruction in code:
            self._decompile_instruction(instruction, 0)

    def _get_opcode(self, opcode: int) -> str:
        if opcode in opcode_translation_table:
            return opcode_translation_table[opcode]
        else:
            return "UNKNOWN"

    def _decompile_instruction(self, instruction: list, indent: int) -> None:
        indent_str = " " * indent
        op = instruction[0]
        if op == Opcode_NEW_VARIABLE:
            print(f"{indent_str}{opcode_translation_table[op]} {instruction[1]}")
        elif op == Opcode_NEW_SCOPE:
            print(f"{indent_str}{opcode_translation_table[op]} {instruction[1]}")
        elif op == Opcode_LOAD_CONSTANT:
            print(f"{indent_str}{opcode_translation_table[op]} {instruction[1]}")
        elif op == Opcode_STORE_VARIABLE:
            var = self._decompile_expression(instruction[1][0])
            expr = self._decompile_expression(instruction[2][0])
            print(f"{indent_str}{opcode_translation_table[op]} {var} = {expr}")
        elif op == Opcode_LEAVE_SCOPE:
            print(f"{indent_str}{opcode_translation_table[op]} {instruction[1]}")
        elif op == Opcode_LOAD_LIBRARY:
            print(f"{indent_str}{opcode_translation_table[op]} {instruction[1]}")
        elif op == Opcode_PRINT:
            print(f"{indent_str}{opcode_translation_table[op]}({self._decompile_args(instruction[2])})")
        elif op == Opcode_EXTERN_FUNCTION_DEFINITION:
            print(f"{indent_str}{opcode_translation_table[op]} {instruction[1]} => ({self._decompile_expression(instruction[2])}) extern {instruction[3]}, {instruction[4]};")
        elif op == Opcode_IF or op == Opcode_ELIF:
            print(f"{indent_str}{opcode_translation_table[op]} {self._decompile_expression(instruction[1])}")
            print(f"{indent_str}{{ ")
            self._decompile_block(instruction[2], indent + 1)
            print(f"{indent_str}}} ")
        elif op == Opcode_CALL:
            print(f"{indent_str}{self._decompile_expression(instruction)}")
        elif op == Opcode_HALT:
            if len(instruction) == 1:
                print(f"{indent_str}{opcode_translation_table[op]}")
            else:
                print(f"{indent_str}{opcode_translation_table[op]} {self._decompile_expression(instruction[1])}")
        elif op == Opcode_RET:
            if len(instruction) == 1:
                print(f"{indent_str}{opcode_translation_table[op]}")
            else:
                print(f"{indent_str}{opcode_translation_table[op]} {self._decompile_expression(instruction[1])}")
        elif op == Opcode_CONTINUE:
            print(f"{indent_str}{opcode_translation_table[op]}")
        elif op == Opcode_BREAK:
            print(f"{indent_str}{opcode_translation_table[op]}")
        elif op == Opcode_FUNCTION_DEFINITION:
            print(f"{indent_str}{opcode_translation_table[op]} {instruction[1]} => ({instruction[2]})")
            print(f"{indent_str}{{ ")
            self._decompile_block(instruction[3], indent + 1)
            print(f"{indent_str}}} ")
        elif op == Opcode_SELECT:
            fields = self.parse_fields(instruction[1])
            from_table = self._decompile_expression(instruction[2]) + ' '
            where = self._decompile_expression(instruction[3])
            if len(where) > 0:
                where = ' where ' + where + ' '
            distinct = instruction[4]
            merge = instruction[5]
            into = 'into ' + self._decompile_expression(instruction[6])
            print(f"{indent_str}{opcode_translation_table[op]} select {'distinct ' if distinct else ''}{fields} from {from_table}{where}{'merge ' if merge else ''}{into}")

        else:
            print(opcode_translation_table[op])
            print(f"we do not know this: {op} {instruction}")
            sys.exit(1)

    def parse_fields(self, fields: dict) -> str:
        result = ""
        for field, asfield in fields.items():
            if field == asfield:
                result += field + ', '
            else:
                result += field + ' as ' + asfield + ', '
        return result[:-2]

    def _decompile_block(self, block: list, indent: int) -> None:
        for instruction in block:
            self._decompile_instruction(instruction, indent)

    def _decompile_args(self, args: list) -> str:
        result = ""
        if len(args) == 1:
            result = self._decompile_expression(args[0])
        elif len(args) > 1:
            for arg in args[:-1]:
                result += f"{self._decompile_expression(arg)}, "
            result += f"{self._decompile_expression(args[-1])}"
        return result
    
    def _decompile_expression_list(self, expr: list) -> str:
        result = ""
        if len(expr) == 1:
            result = self._decompile_expression(expr[0])
        elif len(expr) > 1:
            for e in expr[:-1]:
                result += f"{self._decompile_expression(e)}, "
            result += f"{self._decompile_expression(expr[-1])}"
        return result

    def _decompile_expression(self, expr) -> str:
        if isinstance(expr, list):
            return self._decompile_expression_list(expr)
        if len(expr) == 0:
            return ""
        op = expr[0]
        if op == Opcode_NEW_VARIABLE:
            return f"{opcode_translation_table[op]} {expr[1]}"
        elif op == Opcode_LOAD_CONSTANT:
            return f"{opcode_translation_table[op]} {expr[1]}"
        elif op == Opcode_LOAD_VARIABLE:
            return f"{opcode_translation_table[op]} {expr[1]}"
        elif op == Opcode_ADD :
            lhs = self._decompile_expression(expr[1])
            rhs = self._decompile_expression(expr[2])
            return f"{lhs} + {rhs}"
        elif op == Opcode_SUB :
            lhs = self._decompile_expression(expr[1])
            rhs = self._decompile_expression(expr[2])
            return f"{lhs} - {rhs}"
        elif op == Opcode_MUL :
            lhs = self._decompile_expression(expr[1])
            rhs = self._decompile_expression(expr[2])
            return f"{lhs} * {rhs}"
        elif op == Opcode_DIV :
            lhs = self._decompile_expression(expr[1])
            rhs = self._decompile_expression(expr[2])
            return f"{lhs} / {rhs}"
        elif op == Opcode_MOD :
            lhs = self._decompile_expression(expr[1])
            rhs = self._decompile_expression(expr[2])
            return f"{lhs} % {rhs}"
        elif op == Opcode_POW :
            lhs = self._decompile_expression(expr[1])
            rhs = self._decompile_expression(expr[2])
            return f"{lhs} ** {rhs}"
        elif op == Opcode_GTE :
            lhs = self._decompile_expression(expr[1][0])
            rhs = self._decompile_expression(expr[2][0])
            return f"{lhs} >= {rhs}"
        elif op == Opcode_LTE :
            lhs = self._decompile_expression(expr[1][0])
            rhs = self._decompile_expression(expr[2][0])
            return f"{lhs} <= {rhs}"
        elif op == Opcode_EQ  :
            lhs = self._decompile_expression(expr[1][0])
            rhs = self._decompile_expression(expr[2][0])
            return f"{lhs} == {rhs}"
        elif op == Opcode_NEQ :
            lhs = self._decompile_expression(expr[1][0])
            rhs = self._decompile_expression(expr[2][0])
            return f"{lhs} != {rhs}"
        elif op == Opcode_GT :
            lhs = self._decompile_expression(expr[1][0])
            rhs = self._decompile_expression(expr[2][0])
            return f"{lhs} > {rhs}"
        elif op == Opcode_LT :
            lhs = self._decompile_expression(expr[1][0])
            rhs = self._decompile_expression(expr[2][0])
            return f"{lhs} < {rhs}"
        elif op == Opcode_AND :
            lhs = self._decompile_expression(expr[1][0])
            rhs = self._decompile_expression(expr[2][0])
            return f"{lhs} and {rhs}"
        elif op == Opcode_OR :
            lhs = self._decompile_expression(expr[1][0])
            rhs = self._decompile_expression(expr[2][0])
            return f"{lhs} or {rhs}"
        elif op == Opcode_XOR :
            lhs = self._decompile_expression(expr[1][0])
            rhs = self._decompile_expression(expr[2][0])
            return f"{lhs} xor {rhs}"
        elif op == Opcode_CALL:
            return f"{opcode_translation_table[op]} {expr[1]}({self._decompile_args(expr[3])})"
        else:
            raise Exception(f"Unknown expression {expr}")





#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-11-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the Macal decompiler, it will decompile a .mbc file into bytecode mnenomics.

import argparse
import os
from typing import Optional

from macal.__about__ import __author__, __author_email__, __license__, __version__
from macal.config import SearchPath
from macal.macal_decompiler import MacalDecompiler

from macal.macal_binary_file_io import BINARY_FILE_EXTENSION, LoadBinary


def SetupSearchPath() -> None:
    search_path = [
        os.path.dirname(__file__),
        os.path.join(os.path.dirname(__file__), "lib"),
        os.path.join(os.path.dirname(__file__), "lib", "ext"),
        os.getcwd(),
        os.path.join(os.getcwd(), "lib"),
        os.path.join(os.getcwd(), "lib", "ext"),
    ]
    for path in search_path:
        if path not in SearchPath:
            SearchPath.append(path)


def check_file(file) -> Optional[str]:
    if file is None:
        return
    if not file.endswith(BINARY_FILE_EXTENSION):
        if os.path.exists(f"{file}{BINARY_FILE_EXTENSION}"):
            file = f"{file}{BINARY_FILE_EXTENSION}"
        else:
            raise Exception("Invalid Macal bytecode file.")
    return file


def BuildArgParser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Macal decompiler")
    parser.add_argument("file", nargs="?", type=check_file, help="Macal file")
    parser.add_argument("-s", "--verbose", action="store_true", help="Verbose mode")
    parser.add_argument("-v", "--version", action="version", version=f"Macal 2 interpreter v{__version__}")
    return parser


def Run():
    SetupSearchPath()
    parser = BuildArgParser()
    args = parser.parse_args()
    if args.file is None:
        parser.print_help()
        return
    if args.verbose:
        PrintBanner()
    code = LoadBinary(args.file)
    md = MacalDecompiler()
    md.Decompile(code)


def PrintBanner() -> None:
    print(f"Macal decompiler v{__version__}")
    print(f"Author:  {__author__}")
    print(f"Email:   {__author_email__}")
    print(f"License: {__license__}")
    print()


def PrintTokens(tokens: list) -> None:
    print("Tokens:")
    for token in tokens:
        print(token)
    print()


def PrintAst(ast) -> None:
    print("AST:")
    ast.tree()
    print()


if __name__ == "__main__":
    Run()



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      23-11-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the new Macal Runtime, it interpret/execute a compiled Macal (.mcb) file.

import argparse
import os

from .__about__ import __author__, __author_email__, __license__, __version__
from .config import SearchPath

import pyximport; pyximport.install()
from .macal_vm import MacalVm
from .macal_binary_file_io import BINARY_FILE_EXTENSION, LoadBinary



def SetupSearchPath():
    search_path = [
        os.path.dirname(__file__),
        os.path.join(os.path.dirname(__file__), "lib"),
        os.path.join(os.path.dirname(__file__), "lib", "ext"),
        os.getcwd(),
        os.path.join(os.getcwd(), "lib"),
        os.path.join(os.getcwd(), "lib", "ext"),
    ]
    for path in search_path:
        if path not in SearchPath:
            SearchPath.append(path)


def check_file(file) -> str:
    if file is None:
        return
    if not file.endswith(BINARY_FILE_EXTENSION):
        if os.path.exists(f"{file}{BINARY_FILE_EXTENSION}"):
            file = f"{file}{BINARY_FILE_EXTENSION}"
        else:
            raise Exception("Invalid Macal bytecode file.")
    return file


def BuildArgParser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Macal runtime")
    parser.add_argument("file", nargs="?", type=check_file, help=f"Macal bytecode file ({BINARY_FILE_EXTENSION})")
    parser.add_argument("-v", "--version", action="version", version=f"Macal runtime v{__version__}")
    return parser


def ShowResult(exitCode: int) -> None:
    if exitCode == 0:
        print("Process exits OK")
    else:
        print(f"Process exits with error: {exitCode}")
    exit(exitCode)


def Run():
    SetupSearchPath()
    parser = BuildArgParser()
    args = parser.parse_args()
    if args.file is None:
        parser.print_help()
        return
    code = LoadBinary(args.file)
    vm = MacalVm(verbose=False)
    vm.Execute(code)
    ShowResult(vm.exitcode)


if __name__ == "__main__":
    Run()



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


__version__ = "5.0.0"
__author__ = "Marco Caspers"
__author_email__ = "samadevteam@westcon.com"
__copyright__ = "Copyright 2023 Westcon-Comstor"
__license__ = "MIT License"



# Product:   Macal
# Author:    Marco Caspers
# Date:      16-10-2023
#
#    This library is licensed under the MIT license.
#
#    (c) 2023 Westcon-Comstor
#    (c) 2023 WestconGroup, Inc.
#    (c) 2023 WestconGroup International Limited
#    (c) 2023 WestconGroup EMEA Operations Limited
#    (c) 2023 WestconGroup European Operations Limited
#    (c) 2023 Sama Development Team
#
# Macal time library external functions
#

import time
from datetime import datetime

NUM_SECONDS_FIVE_MINUTES = 300
NUM_SECONDS_ONE_HOUR = 3600
TIME_FORMAT = "%Y%m%d%H%M%S"
ISO_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
ISO_TIME_tzFORMAT = "%Y-%m-%dT%H:%M:%S.0Z"


def DateToUnix(var: str) -> int:
    """Convert a date_string of format YYYYMMDDhhmmss to unix time integer.
    Assumes the date string object is UTC time."""
    dt = datetime.strptime(var, TIME_FORMAT)
    epoch = datetime(1970, 1, 1)
    return int((dt - epoch).total_seconds())


def IsoToUnix(var: str) -> int:
    """Convert a date_string of format %Y-%m-%dT%H:%M:%S.%f to unix time integer.
    Assumes the date string object is in iso format."""
    dt = datetime.strptime(var, ISO_TIME_FORMAT)
    epoch = datetime(1970, 1, 1)
    return int((dt - epoch).total_seconds())


def DateFromUnix(var: int) -> str:
    """Converts time in seconds since UNIX EPOCH to UTC Time format"""
    return time.strftime(TIME_FORMAT, time.gmtime(var))


def IsoFromUnix(var: str) -> None:
    """Converts time in seconds since UNIX EPOCH to UTC Time format"""
    return time.strftime(ISO_TIME_tzFORMAT, time.gmtime(var))


def UtcNow() -> str:
    return datetime.utcnow().strftime("%Y%m%d%H%M%S")


def UtcIsoNow() -> str:
    return "{}Z".format(datetime.utcnow().isoformat())


def IsoNow() -> str:
    return datetime.now().isoformat()


def Now() -> str:
    return datetime.now().strftime("%Y%m%d%H%M%S")


def PerfCounter() -> float:
    return time.perf_counter()



# Product:   Macal
# Author:    Marco Caspers
# Date:      16-10-2023
#
#    This library is licensed under the MIT license.
#
#    (c) 2023 Westcon-Comstor
#    (c) 2023 WestconGroup, Inc.
#    (c) 2023 WestconGroup International Limited
#    (c) 2023 WestconGroup EMEA Operations Limited
#    (c) 2023 WestconGroup European Operations Limited
#    (c) 2023 Sama Development Team
#
# String functions library

from typing import Any

from unidecode import unidecode


def StrLen(arg: str) -> int:
    """Implementation of len function"""
    return len(arg)


def StrLeft(arg: str, length: int) -> str:
    """Implementation of left function"""
    return arg[0:length]


def StrMid(arg: str, offset: int, length: int) -> str:
    """Implementation of mid function"""
    return arg[offset : (offset + length)]


def ToString(arg: Any) -> str:
    """Implementation of toString function"""
    if arg is True:
        return "true"
    elif arg is False:
        return "false"
    elif arg is None or arg == "NIL":
        return "nil"
    elif isinstance(arg, dict):
        vd = {k: "nil" if v is None or v == "NIL" else str(v).lower() if isinstance(v, bool) else v for k, v in arg.items()}
        return f"{vd}"
    elif isinstance(arg, list):
        vl = ["nil" if v is None or v == "NIL" else str(v).lower() if isinstance(v, bool) else v for v in arg]
        return f"{vl}"
    return str(arg)


def StrContains(needle: str, haystack: str) -> bool:
    """Implementation of strContains function"""
    if needle is None or haystack is None:
        return False
    return needle in haystack


def StrReplace(var: str, frm: str, wth: str) -> str:
    """Implementation of strReplace function"""
    if var is None:
        return None
    return var.replace(frm, wth)


def StartsWith(haystack: str, needle: str) -> bool:
    """Implementation of StartsWith function"""
    if haystack is None:
        return False
    return haystack.startswith(needle)


def RemoveNonAscii(text: str) -> str:
    """Implementation of RemoveNonAscii function"""
    if text is None:
        return None
    return unidecode(text)


def ReplaceEx(var: str, repl: Any, by: str) -> str:
    """Implementation of ReplaceEx function"""
    result = var
    if result is not None:
        for ch in repl:
            result = result.replace(ch, by)
    return result


def PadLeft(string: str, char: str, amount: int) -> Any:
    """Implementation of PadLeft function"""
    # this is counter intuitive, but the *just functions in python pad the character on the other
    # end as what their name would imply.
    if string is None:
        string = ""
    if char is None:
        char = " "
    if amount is None:
        amount = 0
    return string.rjust(amount, char)


def PadRight(string: str, char: str, amount: int) -> Any:
    """Implementation of PadRight function"""
    # this is counter intuitive, but the *just functions in python pad the character on the other
    # end as what their name would imply.
    if string is None:
        string = ""
    if char is None:
        char = " "
    if amount is None:
        amount = 0
    return string.ljust(amount, char)


def PadCenter(string: str, char: str, amount: int) -> Any:
    """Implementation of PadCenter function"""
    if string is None:
        string = ""
    if char is None:
        char = " "
    if amount is None:
        amount = 0
    return string.center(amount, char)


def StrToInt(str) -> None:
    """Convert String to Integer"""
    global LastErrorMessage
    try:
        return int(str)
    except Exception as ex:
        LastErrorMessage = f"StrToInt(): {ex}"



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the expression node for the AST.

from __future__ import annotations

from typing import Any, Optional

from .ast_node_expression import AstNodeExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from .macal_conversions import convertToValue
from .bytecode_register import BytecodeRegister


class AstNodeRegisterExpression(AstNodeExpression):
    def __init__(
        self, token: LexToken, register: BytecodeRegister
    ) -> AstNodeRegisterExpression:
        super().__init__(token)
        self.expr_type = AstNodetype.REGISTER_EXPRESSION
        self.register: BytecodeRegister = register

    def __str__(self) -> str:
        return f"AstNodeRegisterExpression({self.register.name})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = "") -> None:
        print(f"{indent}RegisterExpression:")
        print(f"{indent}    {self.register.name}")



# Product:   Macal
# Author:    Marco Caspers
# Date:      16-10-2023
#
#    This library is licensed under the MIT license.
#
#    (c) 2023 Westcon-Comstor
#    (c) 2023 WestconGroup, Inc.
#    (c) 2023 WestconGroup International Limited
#    (c) 2023 WestconGroup EMEA Operations Limited
#    (c) 2023 WestconGroup European Operations Limited
#    (c) 2023 Sama Development Team
#
# Syslog/logging external functions

import logging
import logging.config
import sys
from typing import Any


class SysLog:
    def __init__(self):
        self.__address__ = "localhost"
        self.__port__ = 514

        # REF, modified from:
        #   https://stackoverflow.com/a/19367225

        self.__config__ = {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "verbose": {"format": "%(levelname)s %(module)s P%(process)d T%(thread)d %(message)s"},
                "simple": {"format": "%(levelname)s %(message)s"},
            },
            "handlers": {
                "stdout": {
                    "class": "logging.StreamHandler",
                    "stream": sys.stdout,
                    "formatter": "simple",
                },
                "syslog0": {
                    "class": "logging.handlers.SysLogHandler",
                    "address": "/dev/log",
                    "facility": "local0",
                    "formatter": "simple",
                },
                "syslogR": {
                    "class": "logging.handlers.SysLogHandler",
                    "address": (self.__address__, self.__port__),
                    "formatter": "simple",
                },
            },
            "loggers": {
                "macal-syslog": {
                    "handlers": ["syslog0", "stdout"],
                    "level": logging.DEBUG,
                    "propagate": True,
                },
                "macal-syslog-remote": {
                    "handlers": ["syslogR"],
                    "level": logging.DEBUG,
                    "propagate": False,
                },
            },
        }
        logging.config.dictConfig(self.__config__)
        self.handle = None
        self.debug = None
        self.info = None
        self.warn = None
        self.error = None
        self.critical = None
        self.syslog_enabled = False
        self.remote_enabled = False

    def SysLogInit(self, remote: bool):
        if (remote is True and self.remote_enabled) or (remote is False and self.syslog_enabled):
            raise Exception("Syslog is already initialized!")
        if remote is True:
            self.remote_enabled = True
            self.handle = logging.getLogger("macal-syslog-remote")
            self.debug = self.handle.debug
            self.info = self.handle.info
            self.warn = self.handle.warn
            self.error = self.handle.error
            self.critical = self.handle.critical
        else:
            self.syslog_enabled = True
            self.handle = logging.getLogger("macal-syslog")
            self.debug = self.handle.debug
            self.info = self.handle.info
            self.warn = self.handle.warn
            self.error = self.handle.error
            self.critical = self.handle.critical

    def SysLogSetAddress(self, address, port):
        if self.remote_enabled is True or self.syslog_enabled is True:
            raise Exception("Cannot change syslog configuration after initialization.")
        self.__address__ = address
        self.__port__ = port
        logging.config.dictConfig(self.__config__)


SysLogLocal = SysLog()


def Syslog(level: str, message: str) -> None:
    """Implementation of Syslog function"""
    if level == "debug" and SysLogLocal.debug is not None:
        SysLogLocal.debug(message)
    elif (level == "info" or level == "information") and SysLogLocal.info is not None:
        SysLogLocal.info(message)
    elif (level == "warn" or level == "warning") and SysLogLocal.warn is not None:
        SysLogLocal.warn(message)
    elif level == "error" and SysLogLocal.error is not None:
        SysLogLocal.error(message)
    elif level == "critical" and SysLogLocal.critical is not None:
        SysLogLocal.critical(message)
    else:
        raise Exception(f"Invalid syslog level given: {level}")


def SyslogInit(remote: bool) -> None:
    """Implementation of SysLog init function"""
    SysLogLocal.SysLogInit(remote)


def SyslogSetAddress(address: Any, port: Any) -> None:
    """Implementation of SysLog SetAddress function"""
    SysLogLocal.SysLogSetAddress(address, port)



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# Compile time variable

from __future__ import annotations
from typing import Any
from .ast_nodetype import AstNodetype

class MacalVariable:
    # scope is a NewVMScope, but we can't import it here because of circular imports.
    def __init__(self, name: str, scope) -> MacalVariable:
        self.name: str = name
        self.value_type: AstNodetype = AstNodetype.NIL
        self.value: Any = AstNodetype.NIL
        self.index: list = []
        self.scope = scope
        self.const = False
        self.arg = False

    @property
    def is_const(self) -> bool:
        return self.const
    
    @property
    def is_indexed(self) -> bool:
        return len(self.index) > 0

    def __str__(self) -> str:
        return f"CTVariable({self.name}, {self.value_type}, {self.value}{' indexed' if self.is_indexed else ''}{' const' if self.is_const else ''}{' arg' if self.arg else ''})"

    def __repr__(self) -> str:
        return self.__str__()



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      23-11-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the new Macal Compiler, it will lex, parse, compile the source file and saves the result into a binary file.

import argparse
import os

from .__about__ import __author__, __author_email__, __license__, __version__
from .config import SearchPath
from .macal_compiler import MacalCompiler
from .macal_lexer import Lexer
from .macal_parser import Parser
from .macal_binary_file_io import BINARY_FILE_EXTENSION, SOURCE_FILE_EXTENSION, SaveBinary

def SetupSearchPath():
    search_path = [
        os.path.dirname(__file__),
        os.path.join(os.path.dirname(__file__), "lib"),
        os.path.join(os.path.dirname(__file__), "lib", "ext"),
        os.getcwd(),
        os.path.join(os.getcwd(), "lib"),
        os.path.join(os.getcwd(), "lib", "ext"),
    ]
    for path in search_path:
        if path not in SearchPath:
            SearchPath.append(path)


def check_file(file) -> str:
    if file is None:
        return
    if not file.endswith(BINARY_FILE_EXTENSION) and not file.endswith(SOURCE_FILE_EXTENSION):
        if os.path.exists(f"{file}{SOURCE_FILE_EXTENSION}"):
            file = f"{file}{SOURCE_FILE_EXTENSION}"
        else:
            raise Exception("Invalid Macal source file.")
    return file


def BuildArgParser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Macal compiler")
    parser.add_argument("file", nargs="?", type=check_file, help="Macal file")
    parser.add_argument("-o", "--output", help="Output file")
    parser.add_argument("-r", "--reserved", help="a csv list of reserved variables", default="", required=False, type=str)
    parser.add_argument("-s", "--verbose", action="store_true", help="Verbose mode")
    parser.add_argument("-v", "--version", action="version", version=f"Macal 2 compiler v{__version__}")
    return parser


def LoadSource(fileName) -> str:
    with open(fileName, "r") as f:
        text = f.read()
    return str(text)


def FromSource(args, file: str, output_filename: str) -> None:
    # Lex
    source = LoadSource(file)
    tokens = Lexer(source).tokenize()
    if args.verbose:
        PrintTokens(tokens)

    # Parse
    ast = Parser(tokens, args.file).parse()
    if args.verbose:
        PrintAst(ast)
    
    # Handle reserved variables from the CLI
    reserved = []
    if args.reserved != "":
        reserved = args.reserved.split(",")

    # compile    
    compiler = MacalCompiler(verbose=args.verbose)
    instructions = compiler.compile(ast, reserved)

    # save binary file
    SaveBinary(output_filename, instructions)


def Run():
    SetupSearchPath()
    parser = BuildArgParser()
    args = parser.parse_args()
    if args.file is None:
        parser.print_help()
        return
    output_filename = str(args.output or args.file.replace(SOURCE_FILE_EXTENSION, BINARY_FILE_EXTENSION))
    if args.verbose:
        PrintBanner()
    print(f"Compiling {args.file} to {output_filename}")
    FromSource(args, args.file, output_filename)
    print(f"Compiled to {output_filename}")


def PrintBanner() -> None:
    print(f"Macal compiler v{__version__}")
    print(f"Author:  {__author__}")
    print(f"Email:   {__author_email__}")
    print(f"License: {__license__}")


def PrintTokens(tokens: list) -> None:
    print("Tokens:")
    for token in tokens:
        print(token)
    print()


def PrintAst(ast) -> None:
    print("AST:")
    ast.tree()
    print()


if __name__ == "__main__":
    Run()
